/*******************************************************************************
*
*   testing GLX_EXT_texture_from_pixmap with pure X11/GLX program
*
*   This program is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License version 3,
*   as published by the Free Software Foundation.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*   author: Mirco Müller <macslow@bangang.de>
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/keysymdef.h>
#include <X11/extensions/Xrender.h>
#include <X11/extensions/Xcomposite.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <math.h>

#define WIN_WIDTH   300
#define WIN_HEIGHT  300
#define Z_NEAR      1.0f
#define Z_FAR      50.0f
#define FOVY       20.0f

typedef struct _PixmapTexture
{
	Display*  pDisplay;
	Window    window;
	Pixmap    pixmap;
	GLXPixmap glxpixmap;
	GLenum    iTarget;
	GLuint    uiTextureId;
	GLint     iWidth;
	GLint     iHeight;
	GLfloat   fTop;
	GLfloat   fBottom;
} PixmapTexture;

typedef GLvoid
(*glXBindTexImageEXTFunc) (Display*,
			   GLXDrawable,
			   int,
			   const int*);

typedef GLvoid
(*glXReleaseTexImageEXTFunc) (Display*,
			      GLXDrawable,
			      int);

glXBindTexImageEXTFunc    glXBindTexImageEXT    = 0;
glXReleaseTexImageEXTFunc glXReleaseTexImageEXT = 0;

void
opengl_init_tfp (Display* pDisplay,
		 int      iScreen)
{
	const char* pcDummy = NULL;

	pcDummy = glXQueryExtensionsString (pDisplay, iScreen);

	if (strstr (pcDummy, "GLX_EXT_texture_from_pixmap") != NULL)
	{
		glXBindTexImageEXT = (glXBindTexImageEXTFunc)
			glXGetProcAddress ((GLubyte*) "glXBindTexImageEXT");

		glXReleaseTexImageEXT = (glXReleaseTexImageEXTFunc)
			glXGetProcAddress ((GLubyte*) "glXReleaseTexImageEXT");
	}
	else
		printf ("Unable to find texture-from-pixmap support!\n");
}

void
opengl_print_error (char* pcFuncName)
{
	GLenum error;

	error = glGetError ();

	switch (error)
	{
		case GL_NO_ERROR :
			printf ("%s() caused GL_NO_ERROR\n", pcFuncName);
		break;

		case GL_INVALID_ENUM :
			printf ("%s() caused GL_INVALID_ENUM\n", pcFuncName);
		break;

		case GL_INVALID_VALUE :
			printf ("%s() GL_INVALID_VALUE\n", pcFuncName);
		break;

		case GL_INVALID_OPERATION :
			printf ("%s() GL_INVALID_OPERATION\n", pcFuncName);
		break;

		case GL_STACK_OVERFLOW :
			printf ("%s() GL_STACK_OVERFLOW\n", pcFuncName);
		break;

		case GL_STACK_UNDERFLOW :
			printf ("%s() GL_STACK_UNDERFLOW\n", pcFuncName);
		break;

		case GL_OUT_OF_MEMORY :
			printf ("%s() GL_OUT_OF_MEMORY\n", pcFuncName);
		break;

		case GL_TABLE_TOO_LARGE :
			printf ("%s() GL_TABLE_TOO_LARGE\n", pcFuncName);
		break;
	}
}

void
opengl_change_projection (Bool    bParallel,
			  GLsizei iWidth,
			  GLsizei iHeight)
{
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	if (bParallel)
		glOrtho ((GLdouble) -iWidth / 2, (GLdouble) iWidth / 2,
			 (GLdouble) -iHeight / 2, (GLdouble) iHeight / 2,
			 Z_NEAR, Z_FAR);
	else
		gluPerspective (2.0f * FOVY,
				(GLdouble) iWidth / (GLdouble) iHeight,
				Z_NEAR,
				Z_FAR);

	glMatrixMode (GL_MODELVIEW);
}

void
opengl_init (int iWidth,
	     int iHeight)
{
	/* set some GL-states */
	glClearColor (1.0f, 0.5f, 0.25f, 1.0f);
	glColor4f (0.25f, 0.5f, 1.0f, 1.0f);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_TEXTURE_RECTANGLE_ARB);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, 1);
	glDisable (GL_DEPTH_TEST);
	glLineWidth (2.0f);
	/* setup "camera" */
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	gluLookAt (0.0f, 0.0f, Z_NEAR,
		   0.0f, 0.0f, 0.0f,
		   0.0f, 1.0f, 0.0f);
	glTranslatef (0.0f, 0.0f, -Z_NEAR);
}

void
opengl_resize (int iWidth,
	       int iHeight)
{
	glViewport (0, 0, iWidth, iHeight);
	opengl_change_projection (True, iWidth, iHeight);
	glClear (GL_COLOR_BUFFER_BIT);
}

PixmapTexture*
pixmap_texture_new (Display*     pDisplay,
		    Window       window,
		    GLXFBConfig* pConfig)
{
	PixmapTexture*    pPixmapTexture = NULL;
	int               aiGLXPixmapAttribs[] = {
				GLX_TEXTURE_TARGET_EXT,
				GLX_TEXTURE_RECTANGLE_EXT,
				GLX_TEXTURE_FORMAT_EXT,
				GLX_TEXTURE_FORMAT_RGB_EXT,
				GLX_MIPMAP_TEXTURE_EXT,
				1,
				None };
	Bool              bInvertedY;
	XWindowAttributes attribs;

	pPixmapTexture = (PixmapTexture*) calloc (1, sizeof (PixmapTexture));
	if (!pPixmapTexture)
		return NULL;

	pPixmapTexture->pDisplay  = pDisplay;
	pPixmapTexture->window    = window;
	pPixmapTexture->pixmap    = XCompositeNameWindowPixmap (pDisplay,
								window);
	pPixmapTexture->glxpixmap = glXCreatePixmap (pDisplay,
						     *pConfig,
						     pPixmapTexture->pixmap,
						     aiGLXPixmapAttribs);
	pPixmapTexture->iTarget   = GL_TEXTURE_RECTANGLE_ARB;
	glGenTextures (1, &pPixmapTexture->uiTextureId);

	XGetWindowAttributes (pDisplay,
			      window,
			      &attribs);
	pPixmapTexture->iWidth  = attribs.width;
	pPixmapTexture->iHeight = attribs.height;

	glXGetFBConfigAttrib (pDisplay,
			      *pConfig,
                              GLX_Y_INVERTED_EXT,
                              &bInvertedY);
	if (bInvertedY)
	{
		pPixmapTexture->fTop = 0.0f;
		pPixmapTexture->fBottom = 1.0f * (float)pPixmapTexture->iHeight;
	}
	else
	{
		pPixmapTexture->fTop = 1.0f * (float) pPixmapTexture->iHeight;
		pPixmapTexture->fBottom = 0.0f;
	}

	return pPixmapTexture;
}

void
pixmap_texture_delete (PixmapTexture* pPixmapTexture)
{
	if (!pPixmapTexture)
		return;

	glXDestroyPixmap (pPixmapTexture->pDisplay, pPixmapTexture->glxpixmap);
	XFreePixmap (pPixmapTexture->pDisplay, pPixmapTexture->pixmap);
	glDeleteTextures (1, &pPixmapTexture->uiTextureId);
	free ((void*) pPixmapTexture);
}

void
opengl_draw (Display*       pDisplay,
	     PixmapTexture* pPixmapTexture,
	     GLint          iWidth,
	     GLint          iHeight)
{
	static GLfloat fAngle = 0.0f;

	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* draw orange background with rotating green bar parallel projected */
	opengl_change_projection (True, iWidth, iHeight);
	glPushMatrix ();
	glRotatef (fAngle, 0.0f, 0.0f, 1.0f);
	glBegin (GL_QUADS);
	glColor3f (0.0f, 1.0f, 0.0f);
	glVertex3f (250.0f, 30.0f, 0.0f);
	glVertex3f (-250.0f, 30.0f, 0.0f);
	glVertex3f (-250.0f, -30.0f, 0.0f);
	glVertex3f (250.0f, -30.0f, 0.0f);
	glEnd ();
	glPopMatrix ();
	fAngle += 1.0f;

	/* draw lit textured-cube perspectively projected */
	opengl_change_projection (False, iWidth, iHeight);
	glBindTexture (pPixmapTexture->iTarget,
		       pPixmapTexture->uiTextureId);
	glXBindTexImageEXT (pPixmapTexture->pDisplay,
			    pPixmapTexture->glxpixmap,
			    GLX_FRONT_LEFT_EXT,
			    NULL);
	glTexParameteri (pPixmapTexture->iTarget,
			 GL_TEXTURE_MIN_FILTER,
			 GL_LINEAR);
	glTexParameteri (pPixmapTexture->iTarget,
			 GL_TEXTURE_MAG_FILTER,
			 GL_LINEAR);
	glEnable (GL_DEPTH_TEST);
	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);
	glPushMatrix ();
	glTranslatef ( 1.2f * sin (0.1f * fAngle),
		      -1.5f * cos (0.05f * fAngle),
		      -6.0f + cos (0.15f * fAngle));
	glRotatef (-1.0f * fAngle, 1.0f, 0.0f, 0.0f);
	glRotatef (-0.5f * fAngle, 0.0f, 1.0f, 0.0f);
	glRotatef ( 2.0f * fAngle, 0.0f, 0.0f, 1.0f);
	glBegin (GL_QUADS);
	glNormal3f ( 0.0f, 0.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f, 1.0f, 1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f,-1.0f, 1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f,-1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f, 1.0f,-1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f, 1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f,-1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f,-1.0f,-1.0f);
	glNormal3f ( 1.0f, 0.0f, 0.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f ( 1.0f,-1.0f, 1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f ( 1.0f,-1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f, 1.0f,-1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f (-1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f,-1.0f, 1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f,-1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f (-1.0f, 1.0f,-1.0f);
	glNormal3f ( 0.0f, 1.0f, 0.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f ( 1.0f, 1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f, 1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f (-1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fBottom);
	glVertex3f ( 1.0f,-1.0f, 1.0f);
	glTexCoord2f (0.0f,
		      pPixmapTexture->fTop);
	glVertex3f ( 1.0f,-1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fTop);
	glVertex3f (-1.0f,-1.0f,-1.0f);
	glTexCoord2f ((GLfloat) pPixmapTexture->iWidth,
		      pPixmapTexture->fBottom);
	glVertex3f (-1.0f,-1.0f, 1.0f);
	glEnd ();
	glPopMatrix ();
	glDisable (GL_LIGHTING);
	glDisable (GL_LIGHT0);
	glDisable (GL_DEPTH_TEST);

	glXReleaseTexImageEXT (pPixmapTexture->pDisplay,
			       pPixmapTexture->glxpixmap,
			       GLX_FRONT_LEFT_EXT);

	glXSwapBuffers (pDisplay, glXGetCurrentDrawable ());
}

GLXFBConfig
get_fbconfig (Display* pDisplay,
	      int      iScreen,
	      Window   window,
	      GLfloat* pfTop,
	      GLfloat* pfBottom)
{
	XWindowAttributes attrib;
	VisualID          visualId;
	GLXFBConfig*      pFBConfigs;
	GLXFBConfig       configToUse;
	int               iNumOfConfigs;
	int               i;
	XVisualInfo*      pVisInfo;
	int               iValue;
	int               iTextureTarget = 0;
	int               iDrawableType  = 0;
	int               iBindable      = 0;

	XGetWindowAttributes (pDisplay, window, &attrib);
	visualId = XVisualIDFromVisual (attrib.visual);

	pFBConfigs = glXGetFBConfigs (pDisplay, iScreen, &iNumOfConfigs);

	for (i = 0; i < iNumOfConfigs; i++)
	{
		/* check for matching visual-ids */
		pVisInfo = glXGetVisualFromFBConfig (pDisplay, pFBConfigs[i]);

		if (!pVisInfo || pVisInfo->visualid != visualId)
			continue;
		else
		{
			printf ("\n%d/%d:\n",
				i + 1,
				iNumOfConfigs);

			printf ("  succeeded with visual-id check (depth %d)\n",
				pVisInfo->depth);
		}

		/* check for size of red-component */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_RED_SIZE,
				      &iValue);
		printf ("  red-size      = %2d\n", iValue);

		/* check for size of green-component */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_GREEN_SIZE,
				      &iValue);
		printf ("  green-size    = %2d\n", iValue);

		/* check for size of blue-component */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_BLUE_SIZE,
				      &iValue);
		printf ("  blue-size     = %2d\n", iValue);

		/* check for size of alpha-component */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_ALPHA_SIZE,
				      &iValue);
		printf ("  alpha-size    = %2d\n", iValue);

		/* check for size of buffer */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_BUFFER_SIZE,
				      &iValue);
		printf ("  buffer-size   = %2d\n", iValue);

		/* check for stencil-size */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_STENCIL_SIZE,
				      &iValue);
		printf ("  stencil-size  = %2d\n", iValue);

		/* check for zbuffer-size */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_DEPTH_SIZE,
				      &iValue);
		printf ("  depth-size    = %2d\n", iValue);

		/* check for double-buffering */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_DOUBLEBUFFER,
				      &iValue);
		printf ("  double-buffer = %2d\n", iValue);

		/* check if texture is y-inverted or not */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_Y_INVERTED_EXT,
				      &iValue);
		if (iValue == True)
		{
			*pfTop = 0.0f;
			*pfBottom = 1.0f;
			printf ("  texture is y-inverted\n");
		}
		else
		{
			*pfTop = 1.0f;
			*pfBottom = 0.0f;
			printf ("  texture is not y-inverted\n");
		}

		/* check for bindable color-buffers */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_BIND_TO_TEXTURE_RGB_EXT,
				      &iValue);
		if (iValue)
		{
			iBindable = 1;
			printf ("  color-buffer can be bound to rgb-texture\n");
		}
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_BIND_TO_TEXTURE_RGBA_EXT,
				      &iValue);
		if (iValue)
		{
			iBindable = 1;
			printf ("  color-buffer can be bound to ");
			printf ("rgba-texture\n");
		}
		if (!iBindable)
		{
			printf ("  color-buffer can neither be bound to rgb-");
			printf (" nor to rgba-texture\n");
			continue;
		}

		/* check for window/pixmap/pbuffer drawable-type */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_DRAWABLE_TYPE,
				      &iValue);

		if ((iValue & GLX_WINDOW_BIT))
		{
			iDrawableType = GLX_WINDOW_BIT;
			printf ("  window drawable-type supported\n");
		}
		if ((iValue & GLX_PIXMAP_BIT))
		{
			iDrawableType = GLX_PIXMAP_BIT;
			printf ("  pixmap drawable-type supported\n");
		}
		if ((iValue & GLX_PBUFFER_BIT))
		{
			iDrawableType = GLX_PBUFFER_BIT;
			printf ("  pbuffer drawable-type supported\n");
		}
		if (!iDrawableType)
		{
			printf ("  drawable-type is undetermined\n");
			continue;
		}

		/* check for 1D/2D/rectangular texture-target */
		glXGetFBConfigAttrib (pDisplay,
				      pFBConfigs[i],
				      GLX_BIND_TO_TEXTURE_TARGETS_EXT,
				      &iValue);

		if ((iValue & GLX_TEXTURE_1D_BIT_EXT))
		{
			iTextureTarget = GLX_TEXTURE_1D_BIT_EXT;
			printf ("  1D texture-target supported\n");
		}
		if ((iValue & GLX_TEXTURE_2D_BIT_EXT))
		{
			iTextureTarget = GLX_TEXTURE_2D_BIT_EXT;
			printf ("  2D texture-target supported\n");
		}
		if ((iValue & GLX_TEXTURE_RECTANGLE_BIT_EXT))
		{
			iTextureTarget = GLX_TEXTURE_RECTANGLE_BIT_EXT;
			printf ("  rectangular texture-target supported\n");
		}
		if (!iTextureTarget)
		{
			printf ("  texture-target is undetermined\n");
			continue;
		}

		break;

	}

	if (i == iNumOfConfigs)
		configToUse = None;
	else
		configToUse = pFBConfigs[i];

	return configToUse;
}

Picture
create_picture (Display*      pDisplay,
		Drawable      drawable,
		XRenderColor* pColor,
		int           iWidth,
		int           iHeight,
		Bool          bRepeat,
		int           iDepth,
		int           iFormat)
{
	XRenderPictureAttributes picAttribs;
	unsigned long		 ulMask = 0x00000000;
	Picture			 picture;
	Pixmap			 pixmap;

	pixmap = XCreatePixmap (pDisplay,
				drawable,
				iWidth  <= 0 ? 1 : iWidth,
				iHeight <= 0 ? 1 : iHeight,
				iDepth);

	if (bRepeat)
	{
		picAttribs.repeat = True;
		ulMask            = CPRepeat;
	}

	picture = XRenderCreatePicture (pDisplay,
					pixmap,
					XRenderFindStandardFormat (pDisplay,
								   iFormat),
					ulMask,
					&picAttribs);

	XRenderFillRectangle (pDisplay,
			      PictOpSrc,
			      picture,
			      pColor,
			      0,
			      0,
			      iWidth,
			      iHeight);

	XFreePixmap (pDisplay, pixmap);

	return picture;
}

Picture
create_a8_picture (Display*      pDisplay,
		   Drawable      drawable,
		   XRenderColor* pColor,
		   int           iWidth,
		   int           iHeight,
		   Bool          bRepeat)
{
	return create_picture (pDisplay,
			       drawable,
			       pColor,
			       iWidth,
			       iHeight,
			       bRepeat,
			       8,
			       PictStandardA8);
}

Picture
create_rgb24_picture (Display*      pDisplay,
		      Drawable      drawable,
		      XRenderColor* pColor,
		      int           iWidth,
		      int           iHeight,
		      Bool          bRepeat)
{
	return create_picture (pDisplay,
			       drawable,
			       pColor,
			       iWidth,
			       iHeight,
			       bRepeat,
			       24,
			       PictStandardRGB24);
}

Picture
create_argb32_picture (Display*      pDisplay,
		       Drawable      drawable,
		       XRenderColor* pColor,
		       int           iWidth,
		       int           iHeight,
		       Bool          bRepeat)
{
	return create_picture (pDisplay,
			       drawable,
			       pColor,
			       iWidth,
			       iHeight,
			       bRepeat,
			       32,
			       PictStandardARGB32);
}

void
paint (Display*     pDisplay,
       Window       window,
       Visual*      pVisual,
       int          iWidth,
       int          iHeight,
       GLXFBConfig* pConfig)
{
	XRenderPictFormat*	 pPictFormatA8     = NULL;
	XRenderPictFormat*	 pPictFormatRGB24  = NULL;
	XRenderPictFormat*	 pPictFormatARGB32 = NULL;
	XRenderColor		 colorTrans = {0x0000, 0x0000, 0x0000, 0x0000};
	XRenderColor		 colorWhite = {0xffff, 0xffff, 0xffff, 0xffff};
	Picture			 destPict;
	Picture			 maskPict;
	Picture			 solidPict;
	Picture			 gradientPict;
	XTrapezoid		 trapezoid;
	XLinearGradient		 linearGradient;
	XFixed			 aColorStops[2];
	XRenderColor		 aColorList[2];

	/* get some typical formats */
	pPictFormatA8     = XRenderFindStandardFormat (pDisplay,
						       PictStandardA8);
	pPictFormatRGB24  = XRenderFindStandardFormat (pDisplay, 
						       PictStandardRGB24);
	pPictFormatARGB32 = XRenderFindStandardFormat (pDisplay, 
						       PictStandardARGB32);

	/* this is my target- or destination-drawable derived from the window */
	destPict = XRenderCreatePicture (pDisplay,
					 window,
					 pPictFormatARGB32,
					 0,
					 NULL);

	/* two intermediate pictures */
	solidPict = create_a8_picture (pDisplay,
				       window,
				       &colorWhite,
				       iWidth,
				       iHeight,
				       False);
	maskPict = create_a8_picture (pDisplay,
				      window,
				      &colorTrans,
				      iWidth,
				      iHeight,
				      False);

	/* define a as rectangle/square as trapezoid */
	trapezoid.top        = XDoubleToFixed (0.0f);
	trapezoid.bottom     = XDoubleToFixed ((double) iHeight);
	trapezoid.left.p1.x  = XDoubleToFixed (0.0f);
	trapezoid.left.p1.y  = XDoubleToFixed (0.0f);
	trapezoid.left.p2.x  = XDoubleToFixed (0.0f);
	trapezoid.left.p2.y  = XDoubleToFixed ((double) iHeight);
	trapezoid.right.p1.x = XDoubleToFixed ((double) iWidth);
	trapezoid.right.p1.y = XDoubleToFixed (0.0f);
	trapezoid.right.p2.x = XDoubleToFixed ((double) iWidth);
	trapezoid.right.p2.y = XDoubleToFixed ((double) iHeight);

	/* draw a trapezoid in the mask-picture */
	XRenderCompositeTrapezoids (pDisplay,
				    PictOpOver,
				    solidPict,
				    maskPict,
				    pPictFormatA8,
				    0, 0,
				    &trapezoid,
				    1);

	/* coordinates for the start- and end-point of the linear gradient are
	** in  window-space, they are not normalized like in cairo... here using
	** the height of the window */
	linearGradient.p1.x = XDoubleToFixed (0.0f);
	linearGradient.p1.y = XDoubleToFixed (0.0f);
	linearGradient.p2.x = XDoubleToFixed ((double) iWidth);
	linearGradient.p2.y = XDoubleToFixed (0.0f);

	/* offsets for color-stops have to be stated in normalized form,
	** which means within the range of [0.0, 1.0f] */
	aColorStops[0] = XDoubleToFixed (0.0f);
	aColorStops[1] = XDoubleToFixed (1.0f);

	/* there's nothing much to say about a XRenderColor,
	** each R/G/B/A-component is an unsigned int (16 bit) */
	aColorList[0].red   = 0x0000;
	aColorList[0].green = 0x0000;
	aColorList[0].blue  = 0x0000;
	aColorList[0].alpha = 0xffff;

	aColorList[1].red   = 0x7777;
	aColorList[1].green = 0x7777;
	aColorList[1].blue  = 0x7777;
	aColorList[1].alpha = 0xffff;

	/* create the actual gradient-filling picture */
	gradientPict = XRenderCreateLinearGradient (pDisplay,
						    &linearGradient,
						    aColorStops,
						    aColorList,
						    2);

	/* render a trapezoid */
	XRenderComposite (pDisplay,
			  PictOpSrc,
			  gradientPict,
			  maskPict,
			  destPict,
			  0, 0,
			  0, 0,
			  0, 0,
			  iWidth,
			  iHeight);

	XRenderFreePicture (pDisplay, destPict);
	XRenderFreePicture (pDisplay, solidPict);
	XRenderFreePicture (pDisplay, maskPict);
	XRenderFreePicture (pDisplay, gradientPict);
}

void
list_fbconfigs (Display* pDisplay)
{
	GLXFBConfig*       pConfig             = NULL;
	Bool               bNothing            = True;
	int                iNumOfEntries       = 0;
	int                iEntry              = 0;
	int                iValue              = 0;
	XVisualInfo*       pVisInfo            = NULL;
	XRenderPictFormat* pPictFormat         = NULL;
	int                aiFBConfigAttribs[] = {
		GLX_DRAWABLE_TYPE,               GLX_WINDOW_BIT,
		GLX_RENDER_TYPE,                 GLX_RGBA_BIT,
		GLX_DOUBLEBUFFER,                True,
		GLX_RED_SIZE,                    8,
		GLX_GREEN_SIZE,                  8,
		GLX_BLUE_SIZE,                   8,
		GLX_ALPHA_SIZE,                  8,
		GLX_DEPTH_SIZE,                  24,
		GLX_STENCIL_SIZE,                0,
		GLX_BIND_TO_TEXTURE_RGBA_EXT,    True,
		GLX_BIND_TO_TEXTURE_TARGETS_EXT, GLX_TEXTURE_RECTANGLE_BIT_EXT,
		None};

	pConfig = glXChooseFBConfig (pDisplay,
				     DefaultScreen (pDisplay),
				     aiFBConfigAttribs,
				     &iNumOfEntries);
	printf ("found %d valid entries via glXChooseFBConfig()\n",
		iNumOfEntries);

	for (iEntry = 0; iEntry < iNumOfEntries; iEntry++)
	{
		/* check drawable-type */
		bNothing = True;
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_DRAWABLE_TYPE,
				      &iValue);
		printf ("supported drawable-types: ");
		if ((iValue & GLX_WINDOW_BIT))
		{
			printf ("window");
			bNothing = False;
		}
		if ((iValue & GLX_PIXMAP_BIT))
		{
			printf ("%spixmap", bNothing ? "" : ", ");
			bNothing = False;
		}
		if ((iValue & GLX_PBUFFER_BIT))
		{
			printf ("%spbuffer", bNothing ? "" : ", ");
			bNothing = False;
		}
		if (bNothing)
			printf ("none");
		printf (" (0x%08X)\n", iValue);

		/* check render-type */
		bNothing = True;
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_RENDER_TYPE,
				      &iValue);
		printf ("supported render-types: ");
		if ((iValue & GLX_RGBA_BIT))
		{
			printf ("rgba");
			bNothing = False;
		}
		if ((iValue & GLX_COLOR_INDEX_BIT))
		{
			printf ("%scolor-index", bNothing ? "" : ", ");
			bNothing = False;
		}
		if (bNothing)
			printf ("none\n");
		printf (" (0x%08X)\n", iValue);

		/* check double-buffer support */
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_DOUBLEBUFFER,
				      &iValue);
		printf ("double-buffered: ");
		if (iValue)
			printf ("yes\n");
		else
			printf ("no\n");

		/* check size of red-channel */
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_RED_SIZE,
				      &iValue);
		printf ("red-size: %d\n", iValue);

		/* check size of green-channel */
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_GREEN_SIZE,
				      &iValue);
		printf ("green-size: %d\n", iValue);

		/* check size of blue-channel */
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_BLUE_SIZE,
				      &iValue);
		printf ("blue-size: %d\n", iValue);

		/* check size of alpha-channel */
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_ALPHA_SIZE,
				      &iValue);
		printf ("alpha-size: %d\n", iValue);

		/* check size of depth-buffer */
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_DEPTH_SIZE,
				      &iValue);
		printf ("depth-size: %d\n", iValue);

		/* check size of stencil-buffer */
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_STENCIL_SIZE,
				      &iValue);
		printf ("stencil-size: %d\n", iValue);

		/* check with textures can be used to bind color-buffers */
		bNothing = True;
		printf ("texture-formats supported: ");
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_BIND_TO_TEXTURE_RGB_EXT,
				      &iValue);
		if (iValue)
		{
			printf ("rgb");
			bNothing = False;
		}
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_BIND_TO_TEXTURE_RGBA_EXT,
				      &iValue);
		if (iValue)
		{
			printf ("%srgba", bNothing ? "" : ", ");
			bNothing = False;
		}
		if (bNothing)
			printf ("none");
		printf ("\n");

		/* check the supported texture-targets */
		bNothing = True;
		printf ("texture-targets supported: ");
		glXGetFBConfigAttrib (pDisplay,
				      pConfig[iEntry],
				      GLX_BIND_TO_TEXTURE_TARGETS_EXT,
				      &iValue);
		if ((iValue & GLX_TEXTURE_1D_BIT_EXT))
		{
			printf ("1D");
			bNothing = False;
		}
		if ((iValue & GLX_TEXTURE_2D_BIT_EXT))
		{
			printf ("%s2D", bNothing ? "" : ", ");
			bNothing = False;
		}
		if ((iValue & GLX_TEXTURE_RECTANGLE_BIT_EXT))
		{
			printf ("%sRECT", bNothing ? "" : ", ");
			bNothing = False;
		}
		if (bNothing)
			printf ("none");
		printf (" (0x%08X)\n", iValue);

		/* check for Xrender-compatible visual with alpha-support */
		printf ("Xrender-compatible: ");
		pVisInfo = glXGetVisualFromFBConfig (pDisplay, pConfig[iEntry]);
		if (pVisInfo)
		{
			pPictFormat = XRenderFindVisualFormat (pDisplay,
							       pVisInfo->visual);
			if (pPictFormat)
			{
				printf ("yes");

				if (pPictFormat->direct.alphaMask > 0)
					printf (" (supports alpha-channel)\n");
				else
					printf ("\n");
			}
			else
				printf ("no\n");
		}

		printf ("\n");
	}
}

int 
main (int    argc,
      char** argv)
{
	Display*             pDisplay            = NULL;
	int                  iEventBase          = 0;
	int                  iErrorBase          = 0;
	int                  iMajor              = 0;
	int                  iMinor              = 0;
	XVisualInfo*         pVisInfo            = NULL;
	XVisualInfo          visInfoTemplate;
	int                  iNumOfEntries       = 0;
	int                  i                   = 0;
	XSetWindowAttributes windowAttribs;
	int                  iWindowAttribsMask  = 0;
	Window               window;
	GLXFBConfig          config;
	GLXContext           glxContext;
	GLXWindow            glxWindow;
	float                fTop                = 0.0f;
	float                fBottom             = 0.0f;
	Atom                 property;
	Atom                 type;
	unsigned char        aucWindowTitle[]    = "x11-tfp-test";
	Bool                 bKeepGoing          = True;
	Bool                 bGotEvent           = True;
	XEvent               event;
	int                  iOldWidth           = 0;
	int                  iOldHeight          = 0;
	int                  iNewWidth           = 0;
	int                  iNewHeight          = 0;
	PixmapTexture*       pPixmapTexture      = NULL;

	/* connect to display */
	pDisplay = XOpenDisplay (NULL);
	if (!pDisplay)
	{
		printf ("Could not open display\n");
		return 1;
	}

	/* check for Xcomposite-extension */
	if (!XCompositeQueryExtension (pDisplay, &iEventBase, &iErrorBase))
	{
		XCloseDisplay (pDisplay);
		printf ("Xcomposite-extension not found");
		return 2;
	}

	/* see which version of Xcomposite is available */
	XCompositeQueryVersion (pDisplay, &iMajor, &iMinor);
	if (iMajor == 0 && iMinor < 4)
	{
		XCloseDisplay (pDisplay);
		printf ("At least Xcomposite 0.4 needed\n");
		return 3;
	}
	printf ("Found Xcomposite %d.%d\n", iMajor, iMinor);

	/* check for Xrender-extension */
	if (!XRenderQueryExtension (pDisplay, &iEventBase, &iErrorBase))
	{
		XCloseDisplay (pDisplay);
		printf ("Xrender-extension not found\n");
		return 4;
	}

	/* see which version of Xrender is available */
	XRenderQueryVersion (pDisplay, &iMajor, &iMinor);
	if (iMajor == 0 && iMinor < 10)
	{
		XCloseDisplay (pDisplay);
		printf ("At least Xrender 0.10 needed\n");
		return 5;
	}
	printf ("Found Xrender %d.%d\n", iMajor, iMinor);

	/* check for GLX-extension */
	if (!glXQueryExtension (pDisplay, &iEventBase, &iErrorBase))
	{
		XCloseDisplay (pDisplay);
		printf ("GLX-extension not found\n");
		return 6;
	} 

	/* see which version of GLX is available */
	glXQueryVersion (pDisplay, &iMajor, &iMinor);
	if (iMajor == 1 && iMinor < 2)
	{
		XCloseDisplay (pDisplay);
		printf ("At least GLX 1.2 needed\n");
		return 7;
	}
	printf ("Found GLX %d.%d\n", iMajor, iMinor);

	/* find a rgba-visual */
	visInfoTemplate.depth      = 24;
	visInfoTemplate.red_mask   = 0xff0000;
	visInfoTemplate.green_mask = 0x00ff00;
	visInfoTemplate.blue_mask  = 0x0000ff;
	pVisInfo = XGetVisualInfo (pDisplay,
				   VisualDepthMask  |
				   VisualRedMaskMask|
				   VisualGreenMaskMask|
				   VisualBlueMaskMask,
				   &visInfoTemplate,
				   &iNumOfEntries);
	if (!pVisInfo)
	{
		XCloseDisplay (pDisplay);
		printf ("Failed to obtain VisInfo!\n");
		return 8;
	}
	else
	{
		printf ("Found %d matching Visuals.\n", iNumOfEntries);

		printf ("\nFirst matching visual is...\n");
		printf ("visual-id: 0x%02X\n",
			(int) pVisInfo[0].visualid);
		printf ("  screen:        %d\n",
			pVisInfo[0].screen);
		printf ("  depth:         %d\n",
			pVisInfo[0].depth);
		printf ("  red-mask:      0x%08lx\n",
			pVisInfo[0].red_mask);
		printf ("  green-mask:    0x%08lx\n",
			pVisInfo[0].green_mask);
		printf ("  blue-mask:     0x%08lx\n",
			pVisInfo[0].blue_mask);
		printf ("  colormap-size: %d\n",
			pVisInfo[0].colormap_size);
		printf ("  bits-per-rgb:  %d\n\n",
			pVisInfo[0].bits_per_rgb);
	}

	/* make operations asynchronous */
	XSynchronize (pDisplay, True);

	/* set attribs for the window to be created... and  */
	windowAttribs.border_pixel     = 0;
	windowAttribs.event_mask       = KeyPressMask |
					 StructureNotifyMask |
					 ExposureMask;
	windowAttribs.colormap         = XCreateColormap (
						pDisplay,
						RootWindow (pDisplay,
							    pVisInfo[i].screen),
						pVisInfo[i].visual,
						AllocNone);
	iWindowAttribsMask             = CWBorderPixel |
					 CWEventMask |
					 CWColormap;

	/* create and open window */
	window = XCreateWindow (pDisplay,
				RootWindow (pDisplay, pVisInfo[i].screen),
				50,
				50,
				WIN_WIDTH,
				WIN_HEIGHT,
				0,
				pVisInfo[i].depth,
				InputOutput,
				pVisInfo[i].visual,
				iWindowAttribsMask,
				&windowAttribs);
	XMapWindow (pDisplay, window);

	/* the list of XVisualInfos is no longer needed */
	XFree ((void*) pVisInfo);

	/* redirect window to offscreen-storage */
	XCompositeRedirectWindow (pDisplay,
				  window,
				  CompositeRedirectAutomatic);

	list_fbconfigs (pDisplay);

	/* determine GLXFBConfig for window */
	config = get_fbconfig (pDisplay,
			       DefaultScreen (pDisplay),
			       window,
			       &fTop,
			       &fBottom);
	if (config == None)
	{
		XCompositeUnredirectWindow (pDisplay,
					    window,
					    CompositeRedirectAutomatic);
		XUnmapWindow (pDisplay, window);
		XDestroyWindow (pDisplay, window);
		XCloseDisplay (pDisplay);
		printf ("No suitable suitable GLXFB-config found.\n");
		return 9;
	}

	/* create a GLX context for rendering */
	glxContext = glXCreateNewContext (pDisplay,
					  config,
					  GLX_RGBA_TYPE,
					  NULL,
					  True);

	/* make sure to get a GLX window */
	glxWindow = glXCreateWindow (pDisplay, config, window, NULL);

	/* ensure that GL-context is the one we'll draw to */
	glXMakeContextCurrent (pDisplay, glxWindow, glxWindow, glxContext);

	/* setup of general OpenGL state-variables */
	opengl_init (WIN_WIDTH, WIN_HEIGHT);

	/* this is indeed how you change the window-title under pure X11 */
	property = XInternAtom (pDisplay, "_NET_WM_NAME", True);
	type     = XInternAtom (pDisplay, "UTF8_STRING", True);
	XChangeProperty (pDisplay,
			 window,
			 property,
			 type,
			 8,
			 PropModeReplace,
			 aucWindowTitle,
			 12);

	opengl_init_tfp (pDisplay, DefaultScreen (pDisplay));

	/* event-loop */
	while (bKeepGoing)
	{
		/* check for events _and_ don't block! */
		bGotEvent = XCheckWindowEvent (pDisplay,
					       window,
					       KeyPressMask |
					       StructureNotifyMask |
					       ExposureMask,
					       &event);

		if (bGotEvent)
		{
			switch (event.type)
			{
				case KeyPress :
					if (XLookupKeysym (&event.xkey,
							   0) == XK_Escape)
						bKeepGoing = False;
					if (XLookupKeysym (&event.xkey,
							   0) == XK_q)
						bKeepGoing = False;
				break;

				case ConfigureNotify :
					iNewWidth  = event.xconfigure.width;
					iNewHeight = event.xconfigure.height;

					if ((iOldWidth  != iNewWidth) ||
					    (iOldHeight != iNewHeight))
					{
						opengl_resize (iNewWidth,
							       iNewHeight);
						pixmap_texture_delete (
							pPixmapTexture);
						pPixmapTexture =
						   pixmap_texture_new (pDisplay,
								       window,
								       &config);
					}

					iOldWidth  = iNewWidth;
					iOldHeight = iNewHeight;
				break;

				case Expose :
					/*if (event.xexpose.count < 1)
						paint (pDisplay,
						       window,
						       pVisInfo[i].visual,
						       event.xexpose.width,
						       event.xexpose.height,
						       &config);*/
				break;
			}

		}
		else
		{
			opengl_draw (pDisplay,
				     pPixmapTexture,
				     iOldWidth,
				     iOldHeight);
			usleep (1000000/25);
		}
	}

	/* shutdown, cleanup */
	pixmap_texture_delete (pPixmapTexture);
	glXDestroyWindow (pDisplay, glxWindow);
	glXDestroyContext (pDisplay, glxContext);
	XCompositeUnredirectWindow (pDisplay,
				    window,
				    CompositeRedirectAutomatic);
	XFreeColormap(pDisplay, windowAttribs.colormap);	    
	XUnmapWindow (pDisplay, window);
	XDestroyWindow (pDisplay, window);
	XCloseDisplay (pDisplay);

	return 0;
}

